module Main where

import Prelude hiding (append)

import Control.Monad.Except (runExcept)
import Data.Array as A
import Data.Either (Either(..))
import Data.Int (toNumber)
import Data.Maybe (Maybe(..))
import Data.Traversable (traverse)
import Data.Tuple (Tuple(..))
import DOM.Simple.Console (log, log2)
import Effect (Effect)
import Effect.Aff (launchAff_)
import Effect.Class (liftEffect)
import Effect.Unsafe (unsafePerformEffect)
import Foreign
import Math
import Web.DOM.Document as WDD
import Web.DOM.HTMLCollection as WDH

import Graphics.D3.Base
import Graphics.D3.Contour as Contour
import Graphics.D3.Link as Link
import Graphics.D3.Request
import Graphics.D3.Scale as Scale
import Graphics.D3.Selection as Selection
import Graphics.D3.Selection ((>=>), (>=>-), (>=>++))
import Graphics.D3.SVG.Axis as Axis
import Graphics.D3.Time as Time
import Graphics.D3.Util
import Graphics.D3.Zoom as Zoom

import Example.React as EReact

main :: Effect Unit
main = do
  log2 "d3" d3
  scale <- Scale.linearScale
    >>= Scale.domain [-10.0, 10.0]
    >>= Scale.range [0.0, 1.0]
    >>= Scale.toFunction
  log2 "after scaleLinear" scale
  log2 "scale @ 5" $ scale 5.0

  selected <- Selection.rootSelect "#select-me"
  log2 "selected '#select-me'" selected

  let data' = [ "first data", "second data" ]
  simpleData <- Selection.rootSelect "#simple-data"
    >>= Selection.selectAll "p"
    >>= Selection.bindData data'
    >>= Selection.text' show
  log2 "simpleData" simpleData

  dataWithAppend <- Selection.rootSelect "#data-with-append"
    >>= Selection.selectAll "p"
    >>= Selection.bindData data'
    >>= Selection.enter
    >=>++ "p"
    >>= Selection.style'' "color" (\d idx -> if idx `remainder` 2.0 == 0.0 then "green" else "red")
    >>= Selection.text' show
  log2 "dataWithAppend" dataWithAppend

  -- https://www.d3-graph-gallery.com/graph/density2d_contour.html
  let margin = {top: 20, right: 30, bottom: 30, left: 40}
      width = 460 - margin.left - margin.right
      height = 400 - margin.top - margin.bottom
  contour <- Selection.rootSelect "#contour"
    >=>++ "svg"
    >=> Tuple "width" (width + margin.left + margin.right)
    >=> Tuple "height" (height + margin.top + margin.bottom)
    >=>++ "g"
    >=> Tuple "transform" ("translate(" <> show margin.left <> ", " <> show margin.top <> ")")
  launchAff_ do
    let handleRow d =
          case (runExcept $ traverse readNumber =<< readArray d) of
            Left _ -> { x: 0.0, y: 0.0 }
            Right arr ->
              case A.uncons arr of
                Nothing -> { x: 0.0, y: 0.0 }
                Just { head, tail } -> case A.uncons tail of
                  Nothing -> { x: head, y: 0.0 }
                  Just { head: head2 } -> { x: head, y: head2 }
    data' <- csv "https://raw.githubusercontent.com/holtzy/D3-graph-gallery/master/DATA/data_for_density2d.csv" handleRow
    liftEffect $ log2 "data'" data'
    liftEffect $ do
        x <- Scale.linearScale
          >>= Scale.domain [5.0, 20.0]
          >>= Scale.range [0.0, toNumber width]
        _ <- Selection.append "g" contour
          >=> Tuple "transform" ("translate(0, " <> show height <> ")")
          -- TODO call (d3.axisBottom(x))

        y <- Scale.linearScale
          >>= Scale.domain [5.0, 22.0]
          >>= Scale.range [toNumber height, 0.0]
        _ <- pure contour >>= Selection.append "g"
          -- TODO call (d3.axisLeft(y))

        densityData <- Contour.contourDensity
          >>= Contour.x' (\d -> d.x)
          >>= Contour.y' (\d -> d.y)
          >>= Contour.size [toNumber width, toNumber height]
          >>= Contour.bandwidth 20.0
          >>= Contour.toFunction

        log2 "densityData(data')" $ densityData data'

        _ <- pure contour
          >>= Selection.selectAll "path"
          >>= Selection.bindData (densityData data')
          >>= Selection.enter
          >=>++ "path"
          >=> Tuple "d" Contour.geoPath
          >=> Tuple "fill" "none"
          >=> Tuple "stroke" "#69b3a2"
          >=> Tuple "stroke-linejoin" "round"
        pure unit
  log2 "contour" contour

  -- visualize axis
  axisScale <- Scale.linearScale
    >>= Scale.domain [-1.0, 10.0]
    >>= Scale.range [0.0, 400.0]
  xAxis <- Axis.axisBottom axisScale
  axisSvg <- Selection.rootSelect "#axis"
    >=>++ "svg"
    >=> Tuple "width" 400
    >=> Tuple "height" 50
    >=>++ "g"
    >>= Axis.renderAxis xAxis
  log2 "axisScale" axisScale

  timeScale <- Time.scaleTime
    >>= Scale.range [0.0, 400.0]
  timeAxis <- Axis.axisBottom timeScale
  timeSvg <- Selection.rootSelect "#time-axis"
    >=>++ "svg"
    >=> Tuple "width" 400
    >=> Tuple "height" 50
    >=>++ "g"
    >>= Axis.renderAxis timeAxis
  log2 "timeScale" timeScale

  -- http://using-d3js.com/05_08_links.html
  let linkNodeData = [
      {id: "D3",       position: Tuple 2 0, parentPosition: Tuple 2 0}
    , {id: "Shapes",   position: Tuple 1 1, parentPosition: Tuple 2 0}
    , {id: "Scales",   position: Tuple 3 1, parentPosition: Tuple 2 0}
    , {id: "Links",    position: Tuple 0 2, parentPosition: Tuple 1 1}
    , {id: "Areas",    position: Tuple 1 2, parentPosition: Tuple 1 1}
    , {id: "Arcs",     position: Tuple 2 2, parentPosition: Tuple 1 1}
    , {id: "Ordinal",  position: Tuple 3 2, parentPosition: Tuple 3 1}
    , {id: "Quantize", position: Tuple 4 2, parentPosition: Tuple 3 1}
  ]
  linkXScale <- Scale.linearScale
    >>= Scale.domain [0.0, 4.0]
    >>= Scale.range [25.0, 175.0]
    >>= Scale.toFunction
  linkYScale <- Scale.linearScale
    >>= Scale.domain [0.0, 2.0]
    >>= Scale.range [25.0, 175.0]
    >>= Scale.toFunction
  linkGen <- Link.linkVertical
    >>= Link.source (\d -> d.parentPosition)
    >>= Link.target (\d -> d.position)
    >>= Link.x (\(Tuple x _) -> linkXScale x)
    >>= Link.y (\(Tuple _ y) -> linkYScale y)
  linkElSvg <- Selection.rootSelect "#links"
    >>= Selection.append "svg"
    >=> Tuple "width" 200.0
    >=> Tuple "height" 200.0
  -- _ <- Selection.selectAll ".circle" linkElSvg
  --   >>= Selection.bindData linkNodeData
  --   >>= Selection.join "circle"
  --   >>= Selection.attr' "cx" (\{ position: Tuple x _ } -> linkXScale (toNumber x))
  --   >>= Selection.attr' "cy" (\{ position: Tuple _ y } -> linkYScale (toNumber y))
  --   >>= Selection.attr "r" "5"
  --   >>= Selection.attr "stroke" "black"
  --   >>= Selection.classed "circle" true
  linkG <- Selection.selectAll ".circle" linkElSvg
    >>= Selection.bindData linkNodeData
    >>= Selection.enter
    >>= Selection.append "g"
    -- >>= Selection.attr' "transform" (\{ position: Tuple x y} -> "translate(" <> (show $ linkXScale $ toNumber x) <> ", " <> (show $ linkYScale $ toNumber y) <> ")")
  _ <- Selection.append "circle" linkG
    >=>- Tuple "cx" (\{ position: Tuple x _} -> linkXScale $ toNumber x)
    >=>- Tuple "cy" (\{ position: Tuple _ y} -> linkYScale $ toNumber y)
    >=> Tuple "r" 5
    >=> Tuple "stroke" "black"
    >>= Selection.classed "circle" true
  _ <- Selection.append "text" linkG
    >>= Selection.text' (\{ id } -> id)
    >=>- Tuple "x" (\{ position: Tuple x _} -> linkXScale (toNumber x) + 5.0)
    >=>- Tuple "y" (\{ position: Tuple _ y} -> linkYScale $ toNumber y)
    >=> Tuple "stroke" "black"
    >=> Tuple "font-size" "0.5em"
  _ <- Selection.selectAll "path" linkG
    >>= Selection.bindData linkNodeData
    >>= Selection.join "path"
    >=> Tuple "d" linkGen
    >=> Tuple "fill" "none"
    >=> Tuple "stroke" "black"
  linkZoom <- Zoom.zoom
    >>= Zoom.on "zoom" (\ze -> do
      _ <- log2 "zoom" ze
      _ <- Selection.attr "transform" (Zoom.transform ze) linkG
      pure unit)
  _ <- Zoom.renderZoom linkZoom linkElSvg
  log2 "linkEl" linkElSvg
  resetZoomEl <- Selection.rootSelect "#reset-zoom"
    >>= Selection.onClick (\_ -> do
      _ <- Selection.call (Zoom.transform linkZoom) Zoom.zoomIdentity linkG
      log "zoom reset")
  log2 "resetZoomEl" resetZoomEl

  launchAff_ $ do
    xmlData <- xml "http://localhost:8888/data/employees.xml"
    liftEffect $ log2 "xml" xmlData
    liftEffect $ do
      selXml <- Selection.rootSelect "#xml"
      log2 "selXml" selXml
      rows <- WDD.getElementsByTagName "row" xmlData >>= WDH.toArray
      log2 "rows" rows

  log "xml done"

  EReact.main
